import { onMounted, Ref, ref, watch } from "vue";

type UseRequestResult<R = any, P extends any[] = any[]> = {
    data: Ref<R | undefined>;
    loading: Ref<boolean>;
    error: Ref<Error | undefined>;
    refresh: () => void;
    run: (...args: P) => Promise<void>;
    runAsync: (...args: P) => Promise<void>;
}

export type Service<R, P extends any[]> = (...args: P) => Promise<R>;
export default function useRequest<T = any, P extends any[] = any[]>(
    service: Service<T, P>,
    options?: {
        defaultParams?: P;
        manual?: boolean;
        refreshDeps?: any[];
        onSuccess?: (res: T, params: P) => void;
        onError?: (e: Error, params: P) => void;
    }
): UseRequestResult<T, P> {
    const params = ref<P>((options?.defaultParams || []) as P) as Ref<P>;
    const data = ref<T>();
    const loading = ref(false);
    const error = ref<Error>();

    const _request = (...args: P) => {
        loading.value = true
        return service(...args)
            .then((res) => {
                data.value = res
                options?.onSuccess?.(res, args)
            })
            .catch((e: Error) => {
                error.value = e
                options?.onError?.(e, args)
            })
            .finally(() => {
                loading.value = false
            })
    }
    const runAsync = async (...args: P) => {
        try {
            loading.value = true
            const _ = await service(...args);
            data.value = _
            loading.value = false
        } catch (e) {
            loading.value = false
            error.value = e as Error;
            throw e
        }
    }
    const refresh = () => {
        _request(...params.value)
    }
    onMounted(() => {
        if (!options?.manual) {
            _request(...params.value)
        }
    });
    if (options?.refreshDeps) {
        watch(options.refreshDeps, () => {
            _request(...params.value)
        })
    }
    return {
        data,
        refresh,
        loading,
        error,
        run: _request,
        runAsync
    }
}